/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { ISnapshotTree } from "@fluidframework/protocol-definitions";
import { IGarbageCollectionData, IGarbageCollectionState, IGarbageCollectionDetailsBase, IGarbageCollectionSnapshotData } from "@fluidframework/runtime-definitions";
/**
 * Trims the leading and trailing slashes from the given string.
 * @param str - A string that may contain leading and / or trailing slashes.
 * @returns A new string without leading and trailing slashes.
 */
export declare function trimLeadingAndTrailingSlashes(str: string): string;
/**
 * Trims the leading slashes from the given string.
 * @param str - A string that may contain leading slashes.
 * @returns A new string without leading slashes.
 */
export declare function trimLeadingSlashes(str: string): string;
/**
 * Trims the trailing slashes from the given string.
 * @param str - A string that may contain trailing slashes.
 * @returns A new string without trailing slashes.
 */
export declare function trimTrailingSlashes(str: string): string;
/**
 * Helper function that clones the GC data.
 * @param gcData - The GC data to clone.
 * @returns a clone of the given GC data.
 */
export declare function cloneGCData(gcData: IGarbageCollectionData): IGarbageCollectionData;
/**
 * Helper function that unpacks the GC details of the children from a given node's GC details.
 * @param gcDetails - The GC details of a node.
 * @returns A map of GC details of each children of the the given node.
 */
export declare function unpackChildNodesGCDetails(gcDetails: IGarbageCollectionDetailsBase): Map<string, IGarbageCollectionDetailsBase>;
/**
 * Helper function that unpacks the used routes of children from a given node's used routes.
 * @param usedRoutes - The used routes of a node.
 * @returns A map of used routes of each children of the the given node.
 */
export declare function unpackChildNodesUsedRoutes(usedRoutes: string[]): Map<string, string[]>;
/**
 * Removes the given route from the outbound routes of all the given GC nodes, and any duplicates
 * @param gcNodes - The nodes from which the route is to be removed.
 * @param outboundRoute - The route to be removed.
 */
export declare function removeRouteFromAllNodes(gcNodes: {
    [id: string]: string[];
}, outboundRoute: string): void;
/**
 * Concatenates the given GC states and returns the concatenated GC state.
 */
export declare function concatGarbageCollectionStates(gcState1: IGarbageCollectionState, gcState2: IGarbageCollectionState): IGarbageCollectionState;
/**
 * Concatenates the given GC datas and returns the concatenated GC data.
 */
export declare function concatGarbageCollectionData(gcData1: IGarbageCollectionData, gcData2: IGarbageCollectionData): IGarbageCollectionData;
export declare class GCDataBuilder implements IGarbageCollectionData {
    private readonly gcNodesSet;
    get gcNodes(): {
        [id: string]: string[];
    };
    addNode(id: string, outboundRoutes: string[]): void;
    /**
     * Adds the given GC nodes. It does the following:
     * - Normalizes the ids of the given nodes.
     * - Prefixes the given `prefixId` to the given nodes' ids.
     * - Adds the outbound routes of the nodes against the normalized and prefixed id.
     */
    prefixAndAddNodes(prefixId: string, gcNodes: {
        [id: string]: string[];
    }): void;
    addNodes(gcNodes: {
        [id: string]: string[];
    }): void;
    /**
     * Adds the given outbound route to the outbound routes of all GC nodes.
     */
    addRouteToAllNodes(outboundRoute: string): void;
    getGCData(): IGarbageCollectionData;
}
/**
 * Gets the base garbage collection state from the given snapshot tree. It contains GC state, deleted nodes and
 * tombstones. The GC state may be written into multiple blobs. Merge the GC state from all such blobs into one.
 */
export declare function getGCDataFromSnapshot(gcSnapshotTree: ISnapshotTree, readAndParseBlob: <T>(id: string) => Promise<T>): Promise<IGarbageCollectionSnapshotData>;
//# sourceMappingURL=utils.d.ts.map